﻿//*********************************************************//
//    Copyright (c) Microsoft. All rights reserved.
//    
//    Apache 2.0 License
//    
//    You may obtain a copy of the License at
//    http://www.apache.org/licenses/LICENSE-2.0
//    
//    Unless required by applicable law or agreed to in writing, software 
//    distributed under the License is distributed on an "AS IS" BASIS, 
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or 
//    implied. See the License for the specific language governing 
//    permissions and limitations under the License.
//
//*********************************************************//

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Microsoft.NodejsTools.Analysis;

namespace Microsoft.NodejsTools.Interpreter.Default {
#if FALSE
    class CPythonType : IPythonType, ILocatedMember {
        private readonly string _typeName, _doc;
        private readonly bool _includeInModule;
        private readonly BuiltinTypeId _typeId;
        private readonly CPythonModule _module;
        private readonly List<IPythonType> _bases;
        private readonly IPythonType[] _mro;
        private readonly bool _isBuiltin;
        private readonly Dictionary<string, IMember> _members = new Dictionary<string, IMember>();
        private readonly bool _hasLocation;
        private readonly int _line, _column;

        public CPythonType(IMemberContainer parent, ITypeDatabaseReader typeDb, string typeName, Dictionary<string, object> typeTable, BuiltinTypeId typeId) {
            Debug.Assert(parent is CPythonType || parent is CPythonModule);
            Debug.Assert(!typeId.IsVirtualId());

            _typeName = typeName;
            _typeId = typeId;
            _module = GetDeclaringModule(parent);

            object value;
            if (typeTable.TryGetValue("is_hidden", out value)) {
                _includeInModule = !Convert.ToBoolean(value);
            } else {
                _includeInModule = true;
            }

            if (typeTable.TryGetValue("doc", out value)) {
                _doc = value as string;
            }

            if (typeTable.TryGetValue("builtin", out value)) {
                _isBuiltin = Convert.ToBoolean(value);
            } else {
                _isBuiltin = true;
            }

            if (typeTable.TryGetValue("bases", out value)) {
                var basesList = (List<object>)value;
                if (basesList != null) {
                    _bases = new List<IPythonType>();
                    foreach (var baseType in basesList) {
                        typeDb.LookupType(baseType, StoreBase);
                    }
                }
            }

            if (typeTable.TryGetValue("mro", out value)) {
                var mroList = (List<object>)value;
                if (mroList != null && mroList.Count >= 1) {
                    _mro = new IPythonType[mroList.Count];
                    // Many scraped types have invalid MRO entries because they
                    // report their own module/name incorrectly. Since the first
                    // item in the MRO is always self, we set it now. If the MRO
                    // has a resolvable entry it will replace this one.
                    _mro[0] = this;
                    for (int i = 0; i < mroList.Count; ++i) {
                        var capturedI = i;
                        typeDb.LookupType(mroList[i], t => _mro[capturedI] = t);
                    }
                }
            }

            if (typeTable.TryGetValue("members", out value)) {
                var membersTable = (Dictionary<string, object>)value;
                if (membersTable != null) {
                    LoadMembers(typeDb, membersTable);
                }
            }

            _hasLocation = false;//PythonTypeDatabase.TryGetLocation(typeTable, ref _line, ref _column);
        }

        private CPythonModule GetDeclaringModule(IMemberContainer parent) {
            return parent as CPythonModule ?? (CPythonModule)((CPythonType)parent).DeclaringModule;
        }

        private void StoreBase(IPythonType type) {
            if (type != null && _bases != null) {
                _bases.Add(type);
            }
        }

        private void LoadMembers(ITypeDatabaseReader typeDb, Dictionary<string, object> membersTable) {
            foreach (var memberEntry in membersTable) {
                var memberName = memberEntry.Key;
                var memberValue = memberEntry.Value as Dictionary<string, object>;

                if (memberValue != null) {
                    _members[memberName] = null;
                    typeDb.ReadMember(memberName, memberValue, StoreMember, this);
                }
            }
        }

        private void StoreMember(string memberName, IMember value) {
            _members[memberName] = value;
        }

        public bool IncludeInModule {
            get {
                return _includeInModule;
            }
        }

        #region IPythonType Members

        public IMember GetMember(IModuleContext context, string name) {
            IMember res;
            if (_members.TryGetValue(name, out res)) {
                return res;
            }
            if (_mro != null) {
                foreach (var mroType in _mro.OfType<CPythonType>()) {
                    if (mroType._members.TryGetValue(name, out res)) {
                        return res;
                    }
                }
            } else if (_bases != null) {
                foreach (var baseType in _bases) {
                    res = baseType.GetMember(context, name);
                    if (res != null) {
                        return res;
                    }
                }
            }
            return null;
        }

        public IPythonFunction GetConstructors() {
            IMember member;
            if (_members.TryGetValue("__new__", out member) && member is IPythonFunction && ((IPythonFunction)member).Overloads.Count > 0) {
                return member as IPythonFunction;
            } else if (TypeId != BuiltinTypeId.Object && _members.TryGetValue("__init__", out member)) {
                if (member is CPythonMethodDescriptor) {
                    return ((CPythonMethodDescriptor)member).Function;
                }
                return member as IPythonFunction;
            }
            return null;
        }

        public IList<IPythonType> Mro {
            get {
                return _mro;
            }
        }

        public string Name {
            get {
                if (TypeId != BuiltinTypeId.Unknown) {
                    return _module.TypeDb.GetBuiltinTypeName(TypeId);
                }
                return _typeName;
            }
        }

        public string Documentation {
            get { return _doc ?? ""; }
        }

        public BuiltinTypeId TypeId {
            get {
                return _typeId;
            }
        }

        public IPythonModule DeclaringModule {
            get {
                return _module;
            }
        }

        public bool IsBuiltin {
            get {
                return _isBuiltin;
            }
        }

        #endregion

        #region IMemberContainer Members

        public IEnumerable<string> GetMemberNames(IModuleContext moduleContext) {
            var seen = new HashSet<string>();
            foreach (var key in _members.Keys) {
                if (seen.Add(key)) {
                    yield return key;
                }
            }
            if (_mro != null) {
                foreach (var type in _mro.OfType<CPythonType>()) {
                    foreach (var key in type._members.Keys) {
                        if (seen.Add(key)) {
                            yield return key;
                        }
                    }
                }
            } else if (_bases != null) {
                foreach (var type in _bases) {
                    foreach (var key in type.GetMemberNames(moduleContext)) {
                        if (seen.Add(key)) {
                            yield return key;
                        }
                    }
                }
            }
        }

        #endregion

        #region IMember Members

        public PythonMemberType MemberType {
            get { return PythonMemberType.Class; }
        }

        #endregion

        public override string ToString() {
            return String.Format("CPythonType('{0}')", Name);
        }

        #region ILocatedMember Members

        public IEnumerable<LocationInfo> Locations {
            get {
                if (_hasLocation) {
                    yield return new LocationInfo(_module, _line, _column);
                }
            }
        }

        #endregion
    }
#endif
}
